"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var GameObject_1 = require("./GameObject");
var Mesh_1 = require("./Mesh");
var MeshRender_1 = require("./MeshRender");
var GL_1 = require("./gl/GL");
var Material_1 = require("./Material");
var ShaderFXLibs_1 = require("./shaderfx/ShaderFXLibs");
var Shader_1 = require("./shaderfx/Shader");
var Texture_1 = require("./Texture");
var GLMath_1 = require("./math/GLMath");
var GLTFSceneBuilder = /** @class */ (function () {
    function GLTFSceneBuilder(gltfdata, glctx, shaderlib) {
        this.buffers = {};
        this.buffersDesc = {};
        this.materials = {};
        this.images = {};
        this.m_gltfData = gltfdata;
        this.m_glctx = glctx;
        this.m_shaderfxlib = shaderlib;
        this.m_pbrShader = shaderlib.shaderPbrMetallicRoughness;
        this.m_pbrBlendShader = shaderlib.shaderUnlitTexture;
        this.gltf = gltfdata.gltf;
    }
    GLTFSceneBuilder.prototype.release = function () {
    };
    GLTFSceneBuilder.prototype.createScene = function () {
        var gltf = this.gltf;
        var scenes = gltf.scenes;
        if (scenes == null)
            return null;
        var scene = scenes[gltf.scene];
        var nodes = gltf.nodes;
        var scenenodes = scene.nodes;
        var gsceneobj = new GameObject_1.GameObject();
        for (var i = 0, nodeslen = scenenodes.length; i < nodeslen; i++) {
            var gobj = this.buildNode(nodes, scenenodes[i]);
            if (gobj != null) {
                gobj.transform.parent = gsceneobj.transform;
            }
        }
        var gstrs = gsceneobj.transform;
        gstrs.setScale(GLMath_1.glmath.vec3(0.001, 0.001, 0.001));
        return gsceneobj;
    };
    GLTFSceneBuilder.prototype.buildNode = function (nodes, index) {
        var _node = nodes[index];
        if (_node == null) {
            console.error(nodes);
            console.error(index);
            throw new Error("invalid node");
        }
        var gobj = new GameObject_1.GameObject();
        gobj.name = _node.name;
        if (_node.rotation) {
            gobj.transform.setRotation(new GLMath_1.quat(_node.rotation));
        }
        else if (_node.matrix) {
            //Set matrix
            //TODO
            gobj.transform.localMatrix = new GLMath_1.mat4(_node.matrix);
        }
        //gobj.transform.localScale = glmath.vec3(0.001,0.001,0.001);
        if (_node.mesh) {
            var meshrender = this.getMesh(_node.mesh);
            if (meshrender != null) {
                gobj.render = meshrender;
            }
        }
        var _nodeChildren = _node.children;
        if (_nodeChildren != null && _nodeChildren.length > 0) {
            for (var i = 0, len = _nodeChildren.length; i < len; i++) {
                var g = this.buildNode(nodes, _nodeChildren[i]);
                if (g != null) {
                    g.transform.parent = gobj.transform;
                }
            }
        }
        return gobj;
    };
    GLTFSceneBuilder.prototype.getMesh = function (meshid) {
        var _meshes = this.gltf.meshes;
        if (_meshes == null)
            return null;
        var _mesh = _meshes[meshid];
        if (_mesh == null)
            return null;
        var _primitives = _mesh.primitives;
        if (_primitives == null)
            return null;
        var mesh = new Mesh_1.Mesh();
        mesh.name = _mesh.name;
        //TODO
        var _primitive = _primitives[0];
        var _attribute = _primitive.attributes;
        if (_attribute['NORMAL'] != null) {
            var index = _attribute['NORMAL'];
            var buffer = this.getBuffer(index);
            var desc = this.buffersDesc[index];
            mesh.setNormal(buffer, desc.type, desc.size);
        }
        if (_attribute['POSITION'] != null) {
            var index = _attribute['POSITION'];
            var buffer = this.getBuffer(index);
            var desc = this.buffersDesc[index];
            mesh.setPosition(buffer, desc.type, desc.size);
        }
        if (_attribute['TEXCOORD_0'] != null) {
            var index = _attribute['TEXCOORD_0'];
            var buffer = this.getBuffer(index);
            var desc = this.buffersDesc[index];
            mesh.setUV(buffer, desc.type, desc.size);
        }
        if (_attribute['TANGENT'] != null) {
            //TODO
        }
        //Indices
        {
            var index = _primitive.indices;
            if (index != null) {
                var buffer = this.getBuffer(index);
                var desc = this.buffersDesc[index];
                var mode = _primitive.mode == null ? 4 : _primitive.mode;
                mesh.setIndices(buffer, desc.type, mode);
            }
        }
        var mat = null;
        var matid = _primitive.material;
        if (matid != null) {
            mat = this.getMaterial(_primitive.material);
        }
        var meshrender = new MeshRender_1.MeshRender(mesh, mat);
        return meshrender;
    };
    GLTFSceneBuilder.prototype.getBuffer = function (bufferindex) {
        var buffers = this.buffers;
        var cbuffer = buffers[bufferindex];
        if (cbuffer != null) {
            return cbuffer;
        }
        var _accessors = this.gltf.accessors;
        var _accessor = _accessors[bufferindex];
        var _bufferview = this.gltf.bufferViews[_accessor.bufferView];
        if (_bufferview == null) {
            console.error(_accessor);
            console.error(this.gltf.bufferViews[_accessor.bufferView]);
            throw new Error('buffer view is null');
        }
        var rawBuffer = this.m_gltfData.rawBinary;
        var dataType = _accessor.componentType;
        var dataBuffer = null;
        var dataBufferOffset = 0;
        if (_accessor.byteOffset != null)
            dataBufferOffset += _accessor.byteOffset;
        if (_bufferview.byteOffset != null)
            dataBufferOffset += _bufferview.byteOffset;
        var sizeType = _accessor.type;
        var size = this.getSize(sizeType);
        var componentLength = _accessor.count * size;
        if (dataType == GL_1.GL.FLOAT) {
            dataBuffer = new Float32Array(rawBuffer, dataBufferOffset, componentLength);
        }
        else if (dataType == GL_1.GL.UNSIGNED_INT) {
            dataBuffer = new Uint32Array(rawBuffer, dataBufferOffset, componentLength);
        }
        else if (dataType == GL_1.GL.UNSIGNED_SHORT) {
            dataBuffer = new Uint16Array(rawBuffer, dataBufferOffset, componentLength);
        }
        else {
            throw new Error("buffer datatype not supported." + dataType);
        }
        var totalbyte = componentLength * Mesh_1.MeshBufferUtility.TypeSize(dataType);
        this.buffers[bufferindex] = dataBuffer;
        this.buffersDesc[bufferindex] = new Mesh_1.MeshVertexAttrDesc(dataType, size, totalbyte);
        return dataBuffer;
    };
    GLTFSceneBuilder.prototype.getSize = function (type) {
        switch (type) {
            case "SCALAR":
                return 1;
            case "VEC2":
                return 2;
            case "VEC3":
                return 3;
            case "VEC4":
                return 4;
            case "MAT2":
                return 4;
            case "MAT3":
                return 9;
            case "MAT4":
                return 16;
        }
        throw new Error("invalid type " + type);
    };
    GLTFSceneBuilder.prototype.getMaterial = function (index) {
        var _materials = this.gltf.materials;
        if (_materials == null)
            return null;
        var _material = _materials[index];
        var mat = new Material_1.Material(this.m_pbrShader);
        mat.name = _material.name;
        var shadertags = null;
        var shadertagsOverride = false;
        var matDoubleSided = _material.doubleSided;
        if (matDoubleSided == true) {
            if (shadertags == null)
                shadertags = new Shader_1.ShaderTags();
            shadertags.culling = Shader_1.CullingMode.None;
            shadertagsOverride = true;
        }
        var alphaMode = _material.alphaMode;
        if (alphaMode == "BLEND") {
            if (shadertags == null)
                shadertags = new Shader_1.ShaderTags();
            shadertags.blendOp = Shader_1.BlendOperator.ADD;
            shadertags.queue = Shader_1.RenderQueue.Transparent;
            shadertagsOverride = true;
            mat.setShader(this.m_pbrBlendShader);
        }
        if (shadertagsOverride) {
            if (shadertags.queue == null)
                shadertags.queue = Shader_1.RenderQueue.Opaque;
            mat.shaderTags = shadertags;
        }
        //pbr property
        var _pbrMetallicRoughness = _material.pbrMetallicRoughness;
        if (_pbrMetallicRoughness != null) {
            var _baseCOlorFactor = _pbrMetallicRoughness.baseColorFactor;
            if (_baseCOlorFactor != null) {
                mat.setColor(ShaderFXLibs_1.ShaderFXLibs.SH_PBR_BaseColorFactor, new GLMath_1.vec4(_baseCOlorFactor));
            }
            var _baseColorTexture = _pbrMetallicRoughness.baseColorTexture;
            if (_baseColorTexture != null) {
                var tex = this.getImage(_baseColorTexture.index);
                if (tex != null)
                    mat.setTexture(ShaderFXLibs_1.ShaderFXLibs.SH_PBR_BaseColorTexture, tex);
            }
            var _metallicFactor = _pbrMetallicRoughness.metallicFactor;
            if (_metallicFactor != null) {
                mat.setFloat(ShaderFXLibs_1.ShaderFXLibs.SH_PBR_MetallicFactor, _metallicFactor);
            }
            var _roughnessFactor = _pbrMetallicRoughness.roughnessFactor;
            if (_roughnessFactor != null) {
                mat.setFloat(ShaderFXLibs_1.ShaderFXLibs.SH_PBR_RoughnessFactor, _roughnessFactor);
            }
            var _metallicRoughnessTexture = _pbrMetallicRoughness.metallicRoughnessTexture;
            if (_metallicRoughnessTexture != null) {
                var tex = this.getImage(_metallicRoughnessTexture.index);
                mat.setTexture(ShaderFXLibs_1.ShaderFXLibs.SH_PBR_MetallicRoughnessTexture, tex);
            }
        }
        //emissive property
        var _emissiveFactor = _material.emissiveFactor;
        if (_emissiveFactor != null) {
            mat.setColor(ShaderFXLibs_1.ShaderFXLibs.SH_PBR_EmissiveFactor, GLMath_1.glmath.vec4(_emissiveFactor[0], _emissiveFactor[1], _emissiveFactor[2], 0));
        }
        var _emissiveTexture = _material.emissiveTexture;
        if (_emissiveTexture != null) {
            var tex = this.getImage(_emissiveTexture.index);
            mat.setTexture(ShaderFXLibs_1.ShaderFXLibs.SH_PBR_EmissiveTexture, tex);
        }
        this.materials[index] = mat;
        return mat;
    };
    GLTFSceneBuilder.prototype.getImage = function (index) {
        var img = this.images[index];
        if (img != null) {
            return img;
        }
        var _images = this.gltf.images;
        var _image = _images[index];
        var mime = _image.mimeType;
        if (mime != "image/png" && mime != "image/jpg") {
            throw new Error("invalid mime type " + mime);
        }
        var _bufferview = this.gltf.bufferViews[_image.bufferView];
        if (_bufferview == null) {
            return null;
        }
        var rawBuffer = this.m_gltfData.rawBinary;
        var uint8array = new Uint8Array(rawBuffer, _bufferview.byteOffset, _bufferview.byteLength);
        var texture = Texture_1.Texture.createTextureSync(uint8array, _image.mimeType, this.m_glctx);
        this.images[index] = texture;
        return texture;
    };
    return GLTFSceneBuilder;
}());
exports.GLTFSceneBuilder = GLTFSceneBuilder;
