"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ShaderFXLibs_1 = require("./shaderfx/ShaderFXLibs");
var Shadow_1 = require("./render/Shadow");
var Utility_1 = require("./Utility");
var Texture_1 = require("./Texture");
var Material_1 = require("./Material");
var GLContext_1 = require("./gl/GLContext");
var GraphicsRenderCreateInfo = /** @class */ (function () {
    function GraphicsRenderCreateInfo() {
        this.colorFormat = 0x8058;
        this.depthFormat = 0x81A6;
        this.frameBufferResizeDelay = 250;
    }
    return GraphicsRenderCreateInfo;
}());
exports.GraphicsRenderCreateInfo = GraphicsRenderCreateInfo;
var GraphicsRender = /** @class */ (function () {
    function GraphicsRender(canvas, pipeline, creationInfo) {
        this.shadowConfig = new Shadow_1.ShadowConfig();
        this.pause = false;
        this.m_frameBufferInvalid = false;
        this.m_valid = false;
        this.m_time = 0;
        this.m_dt = 0;
        this.m_resizeDelayter = new Utility_1.Delayter();
        GraphicsRender.globalRender = this;
        this.canvas = canvas;
        if (creationInfo == null) {
            creationInfo = new GraphicsRenderCreateInfo();
            this.m_creationInfo = creationInfo;
        }
        var glctx = GLContext_1.GLContext.createFromCanvas(canvas, {
            antialias: true,
            alpha: false,
            depth: false,
            stencil: false
        });
        this.m_glctx = glctx;
        this.m_shaderFXlib = new ShaderFXLibs_1.ShaderFXLibs(glctx);
        //default texture
        var gl = glctx.gl;
        Material_1.Material.DEF_TEXID_NUM = GraphicsRender.TEXID_DEFAULT_TEX;
        Texture_1.Texture.TEMP_TEXID = gl.TEXTURE2;
        this.m_defaultTexture = Texture_1.Texture.crateEmptyTexture(2, 2, glctx);
        gl.activeTexture(gl.TEXTURE3);
        gl.bindTexture(gl.TEXTURE_2D, this.m_defaultTexture.rawtexture);
        gl.frontFace(gl.CCW);
        this.setPipeline(pipeline);
    }
    Object.defineProperty(GraphicsRender.prototype, "isFrameBufferInvalid", {
        get: function () { return this.m_frameBufferInvalid; },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(GraphicsRender.prototype, "time", {
        get: function () { return this.m_time; },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(GraphicsRender.prototype, "deltaTime", {
        get: function () { return this.m_dt; },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(GraphicsRender.prototype, "pipeline", {
        get: function () {
            return this.m_renderPipeline;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(GraphicsRender.prototype, "shaderLib", {
        get: function () {
            return this.m_shaderFXlib;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(GraphicsRender.prototype, "glctx", {
        get: function () {
            return this.m_glctx;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(GraphicsRender.prototype, "defaultTexture", {
        get: function () {
            return this.m_defaultTexture;
        },
        enumerable: true,
        configurable: true
    });
    GraphicsRender.prototype.setPipeline = function (pipeline) {
        if (pipeline == null)
            return;
        var curpipeline = this.m_renderPipeline;
        if (curpipeline != null) {
            curpipeline.release();
            curpipeline = null;
        }
        pipeline.graphicRender = this;
        pipeline.onSetupRender(this.glctx, this.m_creationInfo);
        this.m_renderPipeline = pipeline;
        this.m_valid = true;
    };
    GraphicsRender.prototype.reload = function () {
        var shaderfxlib = this.m_shaderFXlib;
        if (shaderfxlib != null)
            shaderfxlib.reload();
        var renderpipeline = this.m_renderPipeline;
        if (renderpipeline != null)
            renderpipeline.reload();
    };
    GraphicsRender.prototype.release = function () {
        var renderpipe = this.m_renderPipeline;
        if (renderpipe != null) {
            renderpipe.release();
            this.m_renderPipeline = null;
        }
        this.m_shaderFXlib.release();
        this.m_time = 0.0;
        this.m_valid = false;
    };
    GraphicsRender.prototype.resizeCanvas = function (w, h) {
        var _this = this;
        var canvas = this.canvas;
        if (canvas.width == w && canvas.width == h)
            return;
        if (w <= 0 || h <= 0) {
            this.m_frameBufferInvalid = true;
            return;
        }
        else {
            this.m_frameBufferInvalid = false;
        }
        if (!this.m_valid)
            return;
        var delay = this.m_creationInfo.frameBufferResizeDelay;
        if (delay == 0) {
            canvas.width = w;
            canvas.height = h;
            this.m_renderPipeline.resizeFrameBuffer(w, h);
            return;
        }
        else {
            var delayter = this.m_resizeDelayter;
            delayter.delaytime = delay;
            delayter.emit(function () {
                canvas.width = w;
                canvas.height = h;
                _this.m_renderPipeline.resizeFrameBuffer(w, h);
            });
        }
    };
    /**
     *
     * @param scene
     * @param dt deltaTime /s
     */
    GraphicsRender.prototype.render = function (scene, dt) {
        if (this.pause || this.m_frameBufferInvalid)
            return;
        this.m_time += dt;
        this.m_dt = dt;
        var gl = this.m_glctx.gl;
        // gl.clearColor(0,0,0,1);
        // gl.clear(gl.COLOR_BUFFER_BIT);
        var p = this.pipeline;
        if (p == null)
            return;
        p.exec(scene);
    };
    GraphicsRender.prototype.renderToCanvas = function () {
        this.pipeline.onRenderToCanvas();
    };
    GraphicsRender.TEXID_FB = 0;
    GraphicsRender.TEXID_TEMP = 2;
    GraphicsRender.TEXID_DEFAULT_TEX = 3;
    GraphicsRender.TEXID_SHADER_TEX = [4, 5, 6, 7, 8, 9, 10, 11];
    GraphicsRender.TEXID_SHADOW_MAP = [15, 16, 17, 18];
    return GraphicsRender;
}());
exports.GraphicsRender = GraphicsRender;
